"""This module implements a multivariate Gaussian distribution of a random vector
with INDEPENDENT ELEMENTS, i.e., diagonal covariance matrix,
and extends the scipy.stats implementations by including Bayesian learning.

*** Classes:
GaussianRV: a trainable Gaussian distribution of a random vector with independent elements,
    defined by a Gaussian mean array, and gamma-distributed precision parameters

GaussianGivenPrecisionRV: class for the random mean vector of a GaussianRV object
PrecisionRV: class for the random precision vector of a GaussianRV object

StudentRV: a multivariate Student-t distribution for a vector with independent elements,
    used for predictive distributions derived from a GaussianRV instance.

*** Version History:
* Version 1.0.0:
2023-05-02, allow PrecisionRV.a as 1D array, for potential future extension,
            to allow users to set smaller prior inter-individual threshold variance

* Version 0.7.1:
2022-01-19, StudentRV has own property rng, to allow external seed control,
            GaussianRV.predictive, and GaussianGivenPrecision.predictive ordinary methods
            GaussianRV.predictive, and GaussianGivenPrecision.predictive take input rng

* Version 0.7:
2022-01-02, ensure scalar precision shape parameter for EMA usage,
    deactivate GaussianRV.var, cleanup PrecisionRV.mean_inv
    checked connection with predictive StudentRV

* Version 0.6:
2021-12-03, new GaussianRV.initialize method, explicit assignment of precision params

* EmaCalc version 0.5:
2021-10-22, copied PairedCompCalc -> EmaCalc, unchanged
2021-10-22, modified for use in EmaCalc

* Older applications:
2018-08-10, used in general Bayesian mixture models, and package PairedCompCalc
"""
# **** allow array as PrecisionRV.a parameter ? *******

import numpy as np
import logging
from scipy.special import gammaln, psi

logger = logging.getLogger(__name__)
# logger.setLevel(logging.DEBUG)  # *** TEST


# -------------------------------------------
class GaussianRV:
    """Gaussian distribution of a random 1D (row) array
    with independent elements.
    The probability density function is
    p(x | mu, Lambda) propto prod_d Lambda_d^0.5 exp(- 0.5 (x_d - mu_d)^2 Lambda_d ), where
    mu = (..., mu_d, ...) is the mean array, and
    Lambda=(..., Lambda_d, ...) is the vector of precision values,
        = inverse variance
        = inverse diagonal of covariance matrix.

    To allow Bayesian learning, mu and Lambda are random variables, with
    p(mu, Lambda) = p(mu | Lambda) p(Lambda), where
    p(mu | Lambda) is implemented by a GaussianGivenPrecisionRV instance, and
    p(Lambda) is implemented by a PrecisionRV instance.
    """
    log_2_pi = np.log(2 * np.pi)  # class constant for mean_logpdf calc

    def __init__(self, mean, prec):
        """
        :param mean: GaussianGivenPrecisionRV instance
        :param prec: PrecisionRV instance
        """
        self.mean = mean
        self.prec = prec

    @classmethod
    def initialize(cls,
                   loc,
                   prec_a, prec_b,
                   learned_weight=0.001):
        """Create cls instance with default structure
        :param loc: 1D array-like location vector = mean of mean attribute
        :param prec_a: scalar or 1D array-like precision gamma shape parameter
        :param prec_b: 1D array-like precision gamma inverse-scale parameter
            len(prec_b) == len(loc)
        :param learned_weight: (optional) scalar effective number of observations
            learned_weight = 0. gives a non-informative improper prior density
        """
        prec = PrecisionRV(a=prec_a, b=prec_b)
        mean = GaussianGivenPrecisionRV(loc, learned_weight, prec)
        return cls(mean, prec)

    def __repr__(self):
        property_sep = ',\n\t'
        return (self.__class__.__name__ + '(\n\t'
                + property_sep.join(f'{k}={repr(v)}'
                                    for (k, v) in vars(self).items())
                + ')')

    @property
    def loc(self):
        return self.mean.loc

    @property
    def size(self):
        return len(self.loc)

    def mean_logpdf(self, x):
        """E{ ln pdf( x | self ) }, expectation across all parameters of self
        :param x: 1-dim OR M-dim array or array-like list of sample vectors assumed drawn from self
            x[..., :] = ...-th sample row vector
        :return: scalar or array LL, with
            LL[...] = E{ ln pdf( x[..., :] | self ) }
            LL.shape == x.shape[:-1]

        Arne Leijon, 2018-07-08, seems OK,
        slightly less than self.predictive.logpdf, as expected by Jensen's inequality
        """
        x = np.asarray(x)
        if self.mean.learned_weight <= 0.:
            return np.full(x.shape[:-1], -np.inf)
        z2 = np.dot((x - self.loc)**2, self.prec.mean)
        # = Mahanalobis distance, z2.shape == x.shape[:-1]
        return (- z2 - self.size / self.mean.learned_weight
                + np.sum(self.prec.mean_log) - self.size * self.log_2_pi  # np.log(2 * np.pi)
                ) / 2

    def grad_mean_logpdf(self, x):
        """First derivative of self.mean_logpdf(x) w.r.t x
        :param x: 1-dim OR M-dim array or array-like list of sample vectors assumed drawn from self
            x[..., :] = ...-th sample row vector
        :return: array dLL, with
            dLL[..., i] = d E{ ln pdf( x[..., :] | self ) } / d x[..., i]
            dLL.shape == x.shape
        """
        d = np.asarray(x) - self.loc
        return - d * self.prec.mean

    def relative_entropy(self, othr):
        """Kullback-Leibler divergence between self and othr
        :param othr: single instance of same class as self
        :return: scalar KLdiv(q || p) = E_q{ln q(x) / p(x)},
            where q = self and p = othr
        """
        return (self.mean.relative_entropy(othr.mean) +
                self.prec.relative_entropy(othr.prec))

    def predictive(self, rng=None):
        """Predictive distribution of random vector, integrated over parameters.
        :param rng: (optional) random.Generator object
        :return: rv = single StudentRV instance with independent elements

        Scalar Student pdf(x) propto (1 + (1/df) (x-m)^2 / scale^2 )^(- (df + 1) / 2)
        where scale = sqrt{ (1+beta) self.prec.inv_scale / (beta self.prec.shape)
        and Student df = 2* self.prec.shape
        See Leijon EmaCalc report Appendix, or Leijon JASA PairedComp paper appendix
        """
        beta = self.mean.learned_weight
        return StudentRV(loc=self.loc,
                         scale=np.sqrt(self.prec.b * (1. + beta) / (beta * self.prec.a)),
                         df=2 * self.prec.a,
                         rng=rng)

    def adapt(self, x, x_2, w, prior):  # *** special for EmaCalc, not general ***
        """Update distribution parameters using observed data and prior.
        :param x: 2D array with samples assumed drawn from self.
        :param x_2: 2D array with squared sample elements.
        :param w: 1D array with weights of observed samples
        :param prior: prior conjugate distribution, same class as self
        :return: - KLdiv{self || prior)

        Result: updated internal parameters of self
        Method: Leijon EmaCalc report Appendix: sec:GaussGammaUpdate
        """
        self.mean.adapt(x, w, prior=prior.mean)
        self.prec.adapt(x_2, w, prior=prior.prec,
                        new_mean_2=(self.mean.learned_weight * self.mean.loc**2 -
                                    prior.mean.learned_weight * prior.mean.loc**2))
        ll = - self.relative_entropy(prior)
        if self.mean.learned_weight > 0.5:
            logger.debug(f'comp -KLdiv= {ll:.3f}')
            logger.debug(f'comp.mean.learned_weight = {self.mean.learned_weight:.2f}')
            logger.debug('comp.prec.a = ' + np.array_str(np.asarray(self.prec.a),
                                                         precision=2))
            logger.debug('comp.prec.b = ' + np.array_str(self.prec.b,
                                                         precision=2))
            logger.debug('comp.std = ' + np.array_str(np.sqrt(self.prec.mean_inv()),
                                                      precision=2))
        return ll


# ----------------------------------------------------------------------
class GaussianGivenPrecisionRV:
    """Conditional Gaussian distribution of the mean of a Gaussian random vector,
    given the precision array.
    The probability density function is
    p(mu | Lambda) propto prod_d (beta Lambda_d)^0.5 exp(- 0.5 (mu_d - m_d)^2 beta Lambda_d
    where
    mu is a row vector, sample of random vector self
    m is the location of self,
    beta is the scalar learned_weight property
    Lambda is the precision vector.
    """
    def __init__(self, loc, learned_weight, prec):
        """
        Conditional Gaussian vector, given precision matrix
        :param loc: location vector
        :param learned_weight: scalar effective number of learning data
        :param prec: single PrecisionRV instance
        """
        assert np.isscalar(learned_weight), 'learned_weight must be scalar'
        self.loc = np.asarray(loc)
        self.learned_weight = learned_weight
        self.prec = prec

    def __repr__(self):
        return (self.__class__.__name__ + '(' +
                f'loc= {repr(self.loc)}, ' +
                f'learned_weight= {repr(self.learned_weight)}, ' +
                'prec= prec)')

    def adapt(self, x, w, prior):
        """Update distribution parameters using observed data and prior.
        :param x: 2D array or array-like list of sample vectors assumed drawn from self
        :param w: 1D array of weights
            len(w) == x.shape[0]
        :param prior: prior conjugate distribution, same class as self
        :return: None
        Result: updated internal parameters of self
        """
        m = self.loc  # for debug only
        self.learned_weight = prior.learned_weight + np.sum(w)
        sx = prior.learned_weight * prior.loc + np.dot(w, x)
        self.loc = sx / self.learned_weight
        d = self.loc - m  # update change in location
        if self.learned_weight > 0.5:
            logger.debug('comp loc change: '
                         + np.array_str(d, precision=3))

    def relative_entropy(self, othr):
        """Kullback-Leibler divergence between self and othr
        :param othr: single instance of same class as self
        :return: scalar KLdiv[q || p] = E_q{ln q(x) / p(x)},
            where q = self and p = othr
        """
        d = len(self.loc)
        md = self.loc - othr.loc
        beta_pq_ratio = othr.learned_weight / self.learned_weight
        return (othr.learned_weight * np.dot(md**2, self.prec.mean)
                + d * (beta_pq_ratio - np.log(beta_pq_ratio) - 1.)
                ) / 2

    def predictive(self, rng=None):
        """Predictive distribution of self, integrated over self.prec
        p(mu) = integral p(mu | prec) p(prec) d_prec, where
        p(prec) is represented by the PrecisionRV instance self.prec
        :param rng: (optional) random.Generator object
        :return: rv = single StudentRV instance

        Method: see JASA PairedComp paper Appendix
        see also Leijon EmaCalc doc report,
        re-checked 2022-01-01
        """
        beta = self.learned_weight
        return StudentRV(loc=self.loc,
                         scale=np.sqrt(self.prec.b / (self.prec.a * beta)),
                         df=2 * self.prec.a,
                         rng=rng)


# ---------------------------------------------------------------------------
class PrecisionRV:
    """Distribution of the precision vector Lambda of a Gaussian vector
    The probability density function is
    p(Lambda) = prod_d C_d Lambda_d^(a_d - 1) exp(- b_d Lambda_d), i.e., a gamma density,
        where
        a = scalar (or 1D array) shape parameters
        b = 1D array of inverse-scale parameters
        Lambda.shape == b.shape
        a and b must have broadcast-compatible shapes
        C_d = b_d^a / Gamma(a) is the normalization factor
    """
    def __init__(self, a=0., b=1.):
        """
        :param a: scalar or 1D array-like shape parameter(s)
        :param b: 1D array-like with inverse scale parameter(s)
        """
        # assert np.isscalar(a), 'shape parameter should be scalar for EMA usage'
        try:
            a = np.array(a)
            b = np.array(b)
            test = a / b
        except ValueError as e:
            raise RuntimeError('a and b parameters must be broadcast-compatible and > 0. ' + str(e))
        self.a = a
        self.b = b

    def __repr__(self):
        return self.__class__.__name__ + f'(a= {repr(self.a)}, b= {repr(self.b)})'

    @property
    def size(self):
        return self.mean.size

    @property
    def scale(self):
        return 1./self.b

    @property
    def inv_scale(self):
        return self.b

    @property
    def mean(self):
        """E{self}"""
        return self.a / self.b

    def mean_inv(self):  # *** not needed for EMA ?
        """E{ inv(self) }, where
        inv(self) has an inverse-gamma distribution
        """
        if self.a.ndim > 0:
            m = self.b / np.maximum(self.a - 1, np.finfo(float).eps)
            m[self.a <= 1.] = np.nan
            return m
        elif self.a <= 1.:
            return np.full_like(self.b, np.nan)
        else:
            return self.b / (self.a - 1)

    def mode_inv(self):
        """mode{ inv(self) }, where
        inv(self) has an inverse-gamma distribution
        """
        return self.b / (self.a + 1.)

    @property
    def mean_log(self):
        """E{ ln self } element-wise"""
        return psi(self.a) - np.log(self.b)

    def logpdf(self, x):  # ******* not needed for Ema ?  ********
        """ln pdf(x | self)
        :param x: array or array-like list of 2D arrays
        :return: lp = scalar or array, with
            lp[...] = ln pdf(x[..., :] | self)
            lp.shape == x.shape[:-1]
        """
        bx = self.b * np.asarray(x)
        return np.sum((self.a - 1.) * np.log(bx) - bx
                      + np.log(self.b) - gammaln(self.a),
                      axis=-1)

    def adapt(self, x2, w, prior, new_mean_2):
        """Update distribution parameters using observed data and prior.
        :param x2: 2D array or array-like list of squared samples
            for vectors assumed drawn from distribution with precision == self.
        :param w: 1D array with sample weights
        :param prior: prior conjugate distribution, same class as self
        :param new_mean_2: weighted difference (nu * new_loc^2 - nu' * prior_loc^2)
            where nu is the new sum-weight and nu' is the prior sum-weight
        :return: None

        Result: updated internal parameters of self
        Method: Leijon EmaCalc report: sec:GaussGammaUpdate
        """
        self.a = prior.a + np.sum(w) / 2
        # eq:GammaUpdateA
        self.b = prior.b + (np.dot(w, x2) - new_mean_2) / 2
        # eq:GammaUpdateB

    def relative_entropy(q, p):
        """Kullback-Leibler divergence between PrecisionRV q and p,
        :param p: another instance of same class as self = q
        :return: scalar KLdiv( q || p ) = E{ ln q(x)/p(x) }_q

        Arne Leijon, 2018-07-07 copied from gamma.py 2015-10-16
        """
        pb_div_qb = p.b / q.b
        return np.sum(gammaln(p.a) - gammaln(q.a)
                      - p.a * np.log(pb_div_qb)
                      + (q.a - p.a) * psi(q.a)
                      - q.a * (1. - pb_div_qb)
                      )


class StudentRV:
    """Frozen Student distribution of 1D random vector with INDEPENDENT elements
    generalizing scipy.stats.t for vector-valued random variable
    """
    def __init__(self, df, loc=np.array(0.), scale=np.array(1.),
                 rng=None):
        """Create a StudentRV instance
        :param df: scalar or 1D array-like, degrees of freedom
        :param loc: 1D array or array-like list of location elements
        :param scale: scalar or 1D array or array-like list of scale parameter(s)
            df, loc, and scale must have broadcast-compatible shapes
        :param rng: (optional) random.Generator object
        """
        self.df = np.asarray(df)
        self.loc = np.asarray(loc)
        self.scale = np.asarray(scale)
        if rng is None:
            self.rng = np.random.default_rng()
        else:
            self.rng = rng

    def __repr__(self):
        return (self.__class__.__name__
                + f'(df= {repr(self.df)}, '
                + f'loc= {repr(self.loc)}, '
                + f'scale= {repr(self.scale)})')

    @property
    def size(self):
        return self.loc.size  # len(self.loc)

    @property
    def mean(self):
        if self.df.ndim > 0:
            m = self.loc + 0.  # copy
            m[self.df <= 1.] = np.nan
            return m
        elif self.df > 1.:
            return self.loc
        else:
        # if self.df > 1:
        #     return self.loc
        # else:
            return np.full_like(self.loc, np.nan)

    @property
    def var(self):
        """Variance array"""
        if self.df.ndim > 0:
            v = self.scale ** 2 * self.df / np.maximum(self.df - 2., np.finfo(float).eps)
            v[self.df <= 2.] = np.inf
            v[self.df <= 1.] = np.nan
            return v
        elif self.df > 2:  # scalar df
            return self.scale ** 2 * self.df / (self.df - 2.)
        elif self.df > 1:
            return np.full_like(self.loc, np.inf)
        else:
            return np.full_like(self.loc, np.nan)

    def logpdf(self, x):
        """ln pdf(x | self)
        :param x: array or array-like list of sample vectors
            must be broadcast-compatible with self.loc
        :return: lp = scalar or array of logpdf values
            lp[...] = ln pdf[x[..., :] | self)
            lp.shape == x.shape[:-1]
        Arne Leijon, 2018-07-08, **** checked by comparison to scipy.stats.t
        """
        d = (x - self.loc) / self.scale
        return np.sum(- np.log1p(d**2 / self.df) * (self.df + 1) / 2
                      - np.log(self.scale)
                      + gammaln((self.df + 1) / 2) - gammaln(self.df / 2)
                      - 0.5 * np.log(np.pi * self.df),
                      axis=-1)

    def rvs(self, size=None):
        """Random vectors drawn from self.
        :param size: scalar or tuple with number of sample vectors
        :return: x = array of samples
            x.shape == (*size, self.size)
        """
        if size is None:
            s = self.size
        elif np.isscalar(size):
            s = (size, self.size)
        else:
            s = (*size, self.size)
        # z_sc = scipy_t.rvs(df=self.df, size=s)
        z = self.rng.standard_t(df=self.df, size=s)
        # = standardized samples
        return self.loc + self.scale * z


# ------------------------------------------------- TEST
if __name__ == '__main__':
    from scipy.stats import norm
    from scipy.stats import gamma
    from scipy.stats import t as scipy_t  # ******** skip ?
    import copy

    # --------------------------- Test PrecisionRV
    b = np.array([1., 2., 3.])
    nx = 50

    for a in [0.5, 10., [3., 2., 1.]]:
        g = PrecisionRV(a=a, b=b)
        print(f'\n*** Testing {g}:')
        # x = np.array([gamma(a=a, scale=1/b_i).rvs(size=nx)
        #               for b_i in b]).T
        x = gamma(a=a, scale=1/b).rvs(size=(nx, len(b)))
        print(f'mean= {g.mean}')
        print(f'mean_inv= {g.mean_inv()}')
        print(f'mode_inv= {g.mode_inv()}')
        print(f'gamma samples x[:10]= {x[:10]}')
        print(f'mean(x)= {np.mean(x, axis=0)}')
        print(f'PrecisionRV.logpdf(x)= {g.logpdf(x)}')
        # g_ll = np.array([np.sum([gamma(a=a, scale=1/b_i).logpdf(x_si)
        #                          for (b_i, x_si) in zip(b, x_s)])
        #                  for x_s in x])
        g_ll = np.sum(gamma(a=a, scale=1/b).logpdf(x),
                      axis=-1)
        print(f'scipy gamma.logpdf(x)= {g_ll}')

    # --------------------------- Test StudentRV
    df = 10.
    m = [1., 2., 3.]
    s = [3., 2., 1.]
    for df in [0.5, 1.5, 10., [2., 3., 4.]]:
        st = StudentRV(df=df, loc=m, scale=s)
        print(f'\n*** Testing {st}')
        print(f'mean= {st.mean}')
        print(f'var= {st.var}')
        # scipy_x = np.array([scipy_t.rvs(df=df, loc=m_i, scale=s_i, size=nx)
        #                     for (m_i, s_i) in zip(m, s)]).T
        scipy_x = scipy_t.rvs(df=df, loc=m, scale=s, size=(nx, len(m)))
        print(f'scipy_t samples x[:10]= {scipy_x[:10]}')
        print(f'mean(x)= {np.mean(scipy_x, axis=0)}')

        print(f'StudentRV.rvs() = {st.rvs()}')
        x = st.rvs(size=[nx])
        print(f'StudentRV.rvs(size=[nx]) = x[:10] = {x[:10]}')
        print(f'mean(x)= {np.mean(x, axis=0)}')
        print(f'StudentRV.logpdf(x)= {st.logpdf(x)}')
        # st_ll = np.array([np.sum([scipy_t(df=df, loc=m_i, scale=s_i).logpdf(x_si)
        #                          for (m_i, s_i, x_si) in zip(m, s, x_s)])
        #                  for x_s in x])
        st_ll = np.sum(scipy_t(df=df, loc=m, scale=s).logpdf(x),
                       axis=-1)
        print(f'scipy_t.logpdf(x)= {st_ll}')

    # --------------------------- Test GaussianRV
    scale = np.array([1., 2., 3.])
    prec_a = 1.1
    prec_b = prec_a * scale**2
    g = GaussianRV.initialize(loc=[0., 1., 2.],
                              prec_a=prec_a, prec_b=prec_b,
                              learned_weight=2.01)
    print(f'\n*** Testing {g}')
    print(f'\n* g.predictive = {g.predictive()}')
    print(f'g.predictive.var = \n{g.predictive().var}')
    print(f'g.prec.mean= \n{g.prec.mean}')
    print(f'g.prec.mean_inv= \n{g.prec.mean_inv()}')




"""Main script to run Bayesian analysis of data from an Ecological Momentary Assessment (EMA) study.
This script should be used as a template,
to be copied and modified for any desired analysis.

*** Usage: four main steps, see also explicit template example below

*1: Set up an EmaFrame instance to define experiment and to select input data.
*2: Load a set of collected EMA data into an EmaDataSet instance
*3: Initialize and train an EmaModel instance for observed data.
*4: Display results and save figures and tables to a directory tree.

**  Recommended to run a couple of times with the same data set,
    to see random variability in estimated results.


*** Version history:
* Version 1.0.0:
2023-05-17, EmaModel.initialize(..., n_participants_per_comp, ...) defines initial number of GMM components.
            Parameter max_n_comp no longer used.
2023-04-22, adapted to simplified group-key representation in module ema_data.py:
            Group categories specified in EmaFrame.setup(...)
            NOTE: changed signatures EmaFrame.setup(), EmaDataSet.load(), EmaDataSet.save().

* Version 0.9.5:
2023-03-07, ema_display.EmaDisplaySet include observed and model-predicted grade-count histograms
2023-03-01, timestamp_result if needed to avoid over-writing result files

* Version 0.9.3:
2022-08-14, Each ordinal rating scale may be unique or shared by more than one Attribute
2022-07-30, minor cleanup adapting to minor changes in package modules.

* Version 0.9.1:
2022-04-04, all result tables generated and saved as Pandas.DataFrame instances
2022-03-27, NAP and mean-grades results directly from raw data in ema_data.EmaDataSet
2022-03-21, using Pandas DataFrame format in EmaDataSet, allowing many input file formats

* Version 0.7.1:
2022-01-30, allow seed input to EmaModel.initialize for reproducible random results

* Version 0.7: minor update to include new calculation and display functions

* Version 0.6:
2021-12-08, allow user control of model restriction:
            restrict_attribute: sensory-variable location average forced -> 0.
            restrict_threshold: response-threshold median forced -> 0.

* Version 0.5:
2021-11-22, first functional version
"""
# -------- __main__ check to prevent multiprocessor sub-tasks to re-run this script
if __name__ == '__main__':
    import numpy as np
    from pathlib import Path
    import logging
    import datetime as dt
    import pickle

    from EmaCalc.ema_data import EmaFrame, EmaDataSet
    from EmaCalc.ema_model import EmaModel
    from EmaCalc.ema_display import EmaDisplaySet
    from EmaCalc import ema_logging, __version__
    from EmaCalc.ema_display_format import harmonize_ylim

    # ------------------------ Set up working directory and result logging:
    timestamp_result = True  # New result folder for each run, to prevent over-writing
    # timestamp_result = False  # Repeated runs will over-write existing results

    work_path = Path.home() / 'Documents' / 'EMA_sim'  # or whatever...
    data_path = work_path / 'data'  # to use simulation data generated by run_sim.py
    result_path = work_path / 'result'  # or whatever
    model_file = 'test_ema_model.pkl'  # name of saved model file (if saved)

    if timestamp_result:
        t = dt.datetime.now()
        result_path = result_path.with_name(result_path.name +
                                            f'-{t.year}-{t.month:02}-{t.day:02}-{t.hour:02}{t.minute:02}')

    ema_logging.setup(save_path=result_path,
                      log_file='run_ema_log.txt')  # to save the log file
    logger = logging.getLogger(__name__)
    logger.setLevel(logging.INFO)

    logger.info(f'*** Running EmaCalc version {__version__} ***')

    # ------ 1: Define Experimental Framework: Situations, Attributes, and Groups

    # NOTE: This example uses data generated by template script run_sim.py
    # Edit as needed for any other EMA data source

    situations = {'Phase': ('',),  # only One Test phase with empty label
                  'HA': ('A', 'B'),  # Two Hearing-aid programs
                  'CoSS': [f'C{i}' for i in range(1, 8)],  # Seven CoSS categories
                  }  # Categorical variables
    # NOTE: One situation dimension is always phase, even if only ONE phase category.
    # User may set arbitrary phase_key label.
    # Dimension 'Phase' may be omitted, if only one phase category.
    # Category labels may be strings or integers, which can be used as part of file names.
    # If string labels are intended, avoid strings that can be interpreted as numbers,
    # because Pandas will read them as numbers by default.

    emf = EmaFrame.setup(situations=situations,
                         phase_key='Phase',
                         attributes={'Speech': ['Very Hard',
                                                'Hard',
                                                'Easy',
                                                'Very Easy',
                                                'Perfect'],
                                     'Comfort': ['Bad',
                                                 'Not Good',
                                                 'Not Bad',
                                                 'Good']},
                         groups={'Age': ['young', 'old'],
                                 # 'Gender': ('M', 'F'),
                                 }
                         )
    # NOTE: situations, attributes, and groups keys will be parts of result file names,
    # so they can include only characters allowed in file names.

    # NOTE: situations, attributes, and groups are always analyzed as Categorical variables,
    # even if the categories / grades are defined by numeric labels.
    # The rank order of attribute grades is defined by the order given in EmaFrame.setup(...),
    # NOT by, e.g., alphabetical or numerical order of the grade labels.

    # NOTE: if EQUAL grade labels are used for more than one attribute,
    # the model still assumes separate rating scales for each attribute.
    # However, if grade sequences are IDENTICAL objects for more than one attribute,
    # the model uses the SAME rating scale for those attributes.
    #
    # Example:
    # common_ordinal_scale = ['Very Hard',
    #                         'Hard',
    #                         'Easy',
    #                         'Very Easy',
    #                         'Perfect']
    # emf = EmaFrame.setup(situations=situations,
    #                      phase_key='Phase',
    #                      attributes={'Speech': common_ordinal_scale,  # IDENTICAL scale object for both
    #                                  'Comfort': common_ordinal_scale}
    #                      )

    # In either case, response thresholds are always estimated separately for each participant.

    # ------ 2: Load data from previously saved file(s):

    ds = EmaDataSet.load(emf, data_path, fmt='csv', participant='file',
                         path_groups=['Age',  # find 'Age_old' or 'Age_young' in path string
                                      ],
                         # group_join_str='_'  # default in path-string
                         )
    # NOTE: if 'path_groups' elements are specified, group names in path string must match EXACTLY
    # the labels specified in EmaFrame.setup(...,groups={...}), joined by group_join_str,
    # e.g. with directory data_path / 'Age_old'
    # containing data file(s) for group dimension = 'Age', category = 'old'

    logger.info(f'Using data set ds=\n{ds}')

    # ----------------- (Optional) show rating counts for all participants
    for attr in emf.attribute_dtypes.keys():
        a_count = ds.attribute_grade_count(attr, groupby='HA')
        logger.info(str(attr) + '_grade_count:\n' + a_count.to_string())
        logger.info(str(attr) + ': sum grade_count= ' + f'{np.sum(a_count.to_numpy())}')
        # a_count.save(result_path / (str(attr) + '_grades.csv'))

    # ----------------- (Optional) show mean grades and NAP results for all participants
    mean_grades = ds.attribute_grade_mean(groupby=('HA', 'CoSS'))
    mean_grades.save(result_path / 'Attribute_mean_grades.txt', float_format='%.2f')
    # mean_grades.save(result_path / 'Attribute_mean_grades.csv',
    #                  float_format='%.4f')  # if needed for other analysis
    logger.info(f'Attribute mean grades saved in {result_path}'
                + '\nCAUTION: Mean values presume METRIC data, but attribute grades are only ORDINAL!')

    nap = ds.nap_table('HA', nap_cat=['A', 'B'], p=0.95)  # aggregated across other situation dimensions
    # nap = ds.nap_table('HA', nap_cat=['A', 'B'], groupby=('CoSS',), p=0.95)  # grouped results
    nap.save(result_path / 'NAP.txt', float_format='%.2f')  # pretty-formatted by Pandas
    # nap.save(result_path / 'NAP.tex', float_format='%.2f')  # for import to LaTeX doc.
    # nap.save(result_path / 'NAP.csv', sep='\t')  # tab-delimited text, for input to other program
    logger.info(f'NAP results saved in {result_path}')

    # ------ 3: Learn Analysis Model from loaded data set:

    # Model ordinal-regression effects of Situations on each Attribute:

    # regression_effects = ['HA',     # main linear regression effect only
    #                       'CoSS',   # main linear regression effect only
    #                       # 'Phase',  # if there are several phase categories
    #                       ]

    regression_effects = [('HA', 'CoSS')  # joint effects, main AND interaction
                          # 'Phase',  # if there are several phase categories
                          ]

    # NOTE: A regression_effects element may include any combination of situation dimensions, BUT
    # including ALL interactions -> many model parameters,
    # possibly -> less reliable estimation for each parameter.

    # In this example: ['HA', 'CoSS'] -> 2 + (7 - 1) = 8 regression-effect parameters
    #                ['CoSS', 'HA'] -> 7 + (2 - 1) = 8 regression-effect parameters
    #                [('HA', 'CoSS')] -> 2 * 7 = 14 regression-effect parameters

    emm = EmaModel.initialize(ds,
                              effects=regression_effects,
                              # n_participants_per_comp=5,  # default
                              # restrict_attribute=False,  # default
                              # restrict_threshold=True,  # default
                              # seed=12345  # ONLY if reproducible results are required
                              )
    # n_participants_per_comp = initial number of participants per mixture component in population model
    # restrict_attribute=True -> force attribute mean location at zero
    # restrict_threshold=True -> force one mid-scale response threshold at zero
    # for each respondent and each sample of each attribute.

    ll = emm.learn(max_hours=1., max_minutes=0.)
    logger.info(f'*** Data log-likelihood = {ll[-1]:.1f}, indicating model fit to data. ***')
    # *** Recommended to re-run model learning with same data set a couple of times,
    # and then use best-fitting model result.

    emm.prune()  # keep only active mixture components.

    # -------- Save learned EmaModel (optional):
    with (work_path / model_file).open('wb') as f:
        pickle.dump(emm, f)
        logger.info('Model pickle-dumped as ' + f.name)

    # ------ 4: Generate result displays:

    # -------- Re-load learned EmaModel (optional, if saved):
    # with (work_path / model_file).open('rb') as f:
    #     emm = pickle.load(f)
    # -------------------------------------------------
    emd = EmaDisplaySet.show(emm,
                             situations=['CoSS',  # CoSS probabilities, aggregated across HA
                                         ('CoSS', 'HA'),  # CoSS probabilities, conditional on HA
                                         ('HA', 'CoSS'),  # HA probabilities, conditional on CoSS
                                         ],
                             attributes=[('Speech', 'CoSS'),    # Speech, main effect of CoSS
                                         ('Speech', 'HA'),      # Speech, main effect of HA
                                         ('Speech', ('CoSS', 'HA')),    # joint effect of both
                                         ('Comfort', ('CoSS', 'HA'))],  # joint effect of both
                             grade_counts=['Speech',  # total 'Speech' grade-counts, sum across HA and CoSS
                                           ('Speech', 'HA'),    # 'Speech' grade-counts, separated by HA
                                           ('Comfort', 'HA')],
                             random_individual=True,    # random individual in population
                             population_mean=True,      # population mean
                             participants=False,        # individual results: True -> MANY plots and tables
                             grade_thresholds=True,     # True -> median response thresholds in attribute plots
                             percentiles=[2.5, 25, 50, 75, 97.5],   # in profile plots and tables
                             credibility_limit=0.7,     # minimum credibility in difference tables
                             mpl_params={'figure.max_open_warning': 0,
                                         'axes.labelsize': 'x-large'},  # -> matplotlib.rcParam
                             # mpl_style='my_style_sheet',
                             # ... any other ema_display.FMT or ema_display_format.FMT settings
                             )
    # NOTE: joint (=interaction) effects are correct only if included in model regression_effects

    # ------------------------------- (optionally) edit display elements, if desired
    for g_disp in emd.groups.values():
        harmonize_ylim([g_disp.population_mean.attributes[('Speech', ('CoSS', 'HA'))].plot.ax,
                        g_disp.random_individual.attributes[('Speech', ('CoSS', 'HA'))].plot.ax,
                        g_disp.population_mean.attributes[('Comfort', ('CoSS', 'HA'))].plot.ax,
                        g_disp.random_individual.attributes[('Comfort', ('CoSS', 'HA'))].plot.ax
                        ])
        harmonize_ylim([g_disp.population_mean.situations[('CoSS', 'HA')].plot.ax,
                        g_disp.random_individual.situations[('CoSS', 'HA')].plot.ax,
                        ])
    # -> matching y-axis limits: nice for plots to be shown side by side

    # ------------------------------- save all result displays
    emd.save(result_path, figure_format='pdf', table_format='txt', float_format='%.2f')
    # (optionally) save in other format(s), too:
    # emd.save(result_path,
    #          table_format='csv',  # for input to other package
    #          float_format='%.4f',  # any other parameters for Pandas table-writer function
    #          # sep='\t'  # -> tab-delimited
    #          )

    logging.info(f'All results saved in {result_path}')

    logging.shutdown()



"""This module defines functions that read / write EMA data and analysis result tables,
from/to a table file of any kind that package Pandas can handle.

A single data file for READING may include EMA records from ONE or SEVERAL respondents,
in ONE or SEVERAL respondent groups.
The participant id may be stored in a designated column of the table,
otherwise the file name can be used as participant id,
or (in Excel-type file) the participant may be identified by the sheet name.
The table may include grouping categories, and other data not used for EMA analysis.

Regardless of the file storage format, even if an input file represents only ONE participant,
the EMA data is always delivered by a generator function
ema_gen, which is an iterable over pandas DataFrame instances.

Some pandas file readers can also do some
data conversions and/or type checking.
Column names may be modified before the data are delivered to caller.

For writing EMA data or any other table to a file,
cast the DataFrame instance to subclass Table,
and use Table(df).save(...) method.


*** Version History:
* Version 1.0.0:
2023-04-20, simplified ema_gen(..): yield whole table as a single pd.DataFrame,
            All data checking, and separation by group(s) and participant, is done externally.

* Version 0.9.3:
2022-07-22, changed 'subject' -> 'participant'
2022-07-15, minor bugfix in ema_gen.separate

* Version 0.9:
2022-03-21, use pd.DataFrame as interface to all file formats

* Version 0.8.3 and earlier: Handled only xlsx files.
"""
import pandas as pd
import logging

logger = logging.getLogger(__name__)
# logger.setLevel(logging.DEBUG)


# ------------------------------------------------ File exceptions
class FileReadError(RuntimeError):
    """Any type of exception while attempting to read EMA data from a file
    """


class FileWriteError(RuntimeError):
    """Any type of exception while attempting to write a pd.DataFrame table to a file
    """


# --------------------------------------------------
class Table(pd.DataFrame):
    """Subclass adding a general save method,
    automatically switching to desired file format.
    """
    def save(self, file_path,
             allow_over_write=True,
             write_fcn=None,
             **file_kwargs):
        """Save self to a table-style file.
        :param file_path: Path instance defining file location and full name incl. suffix.
        :param allow_over_write: (optional) boolean, if False, find new unused path stem.
        :param write_fcn: (optional) user-supplied function with signature
            write_fcn(table, path, **kwargs).
            If None, default pd.DataFrame method is used, determined by file_path.suffix
        :param file_kwargs: (optional) any additional arguments for the
            write function for the specific file format.
        :return: None
        """
        if not allow_over_write:
            file_path = safe_file_path(file_path)
        suffix = file_path.suffix
        try:
            # ******** file_path.parent.mkdir here? NO, done by container class
            if write_fcn is None:
                if suffix in ['.xlsx', '.xls', '.odf', '.ods', '.odt']:
                    self.to_excel(file_path, sheet_name=file_path.stem, **file_kwargs)
                elif suffix in ['.csv']:
                    self.to_csv(file_path, **file_kwargs)
                elif suffix in ['.txt']:
                    self.to_string(file_path, **file_kwargs)
                elif suffix in ['.tex']:
                    # with warnings.catch_warnings():
                    #     # suppress Pandas FutureWarning about to_latex method
                    #     warnings.simplefilter('ignore')
                    #     self.to_latex(file_path, **file_kwargs)
                    self.to_latex(file_path, **file_kwargs)
                else:
                    raise FileWriteError(f'No DataFrame write method for file type {suffix}')
            else:
                write_fcn(self, file_path, **file_kwargs)
        except Exception as e:
            raise FileWriteError(f'Could not write to {file_path}. Error: {e}')


# ----------------------------------------- General File Reader:
def ema_gen(file_path,
            participant=None,
            rename_cols=None,
            read_fcn=None,
            **file_kwargs
            ):
    """Reader of data stored in a table-style file that can be read by Pandas.read_xxx()
    or user-supplied function with similar signature.
    Each file row is assumed to include fields for one EMA record, but may also include other data fields.
    :param file_path: Path to existing file for reading
    :param participant: (optional) 'sheet' or 'file' or column header for participant identification label
        None -> 'sheet', if Excel-type file, otherwise -> 'file'
    :param rename_cols: (optional) dict with elements (old_name: new_name) for table column(s)
    :param read_fcn: (optional) basic read function, e.g., pd.read_excel,
        or any user-supplied function with similar signature.
        If None, a default pandas function is determined by file_path.suffix.
    :param file_kwargs: any additional arguments for the
        read function for the specific file format
    :return: generator object yielding pd.DataFrame instance(s),
        with one row for each EMA record,
        and one column specifies participant id,
        possibly copied from file name or sheet name if requested.
    """
    if read_fcn is None:
        read_fcn = _default_reader(file_path)
    logger.debug(f'Reading from {file_path} with {read_fcn}')
    if participant == 'sheet':
        if read_fcn is pd.read_excel:
            if 'sheet_name' not in file_kwargs.keys():
                file_kwargs['sheet_name'] = None
                # because pd.read_excel uses sheet_name=0 by default
        else:
            logger.warning(f'File type {file_path.suffix} has no "sheet"s. Using participant="file".')
            participant = 'file'
    try:
        df = read_fcn(file_path,
                      **file_kwargs)
    except Exception as e:
        raise FileReadError(f'{read_fcn.__name__}({file_path}) error: {e}')
    if type(df) is pd.DataFrame:
        if rename_cols is not None:
            df.rename(columns=rename_cols, inplace=True)
        if participant is None or participant == 'file':
            df['file'] = file_path.stem
            yield df
        elif participant in df.columns:
            yield df
        else:
            logger.debug(f'No column participant={participant} in this file')
    else:  # we have a dict of sheet DataFrames
        for (sheet_name, df_s) in df.items():
            if rename_cols is not None:
                df_s.rename(columns=rename_cols, inplace=True)
            if participant is None or participant == 'sheet':
                df_s['sheet'] = sheet_name
                yield df_s
            elif participant == 'file':
                df_s[participant] = file_path.stem
                yield df_s
            elif participant in df_s.columns:
                yield df_s
            else:
                logger.debug(f'Missing column participant={participant} in sheet {sheet_name} of this file')


# ------------------------------------------ help functions:
def safe_file_path(p):
    """Ensure previously non-existing file path, to avoid over-writing,
    by adding a sequence number to the path stem
    :param p: file path
    :return: p_safe = similar file path with modified name
    """
    f_stem = p.stem
    f_suffix = p.suffix
    i = 0
    while p.exists():
        i += 1
        p = p.with_name(f_stem + f'-{i}' + f_suffix)
    return p


def _default_reader(file_path):
    """Find a function that can read the given file path
    :param file_path: a Path instance
    :return: a pandas reader function, e.g., pd.read_excel
    """
    suffix = file_path.suffix
    if suffix in ['.xlsx', '.xls', '.odf', '.ods', '.odt']:
        return pd.read_excel
    elif suffix in ['.csv', '.txt']:
        return pd.read_csv
    elif suffix in ['.sav', '.zsav']:
        return pd.read_spss
    elif suffix in ['.dta']:
        return pd.read_stata
    else:
        raise FileReadError(f'Cannot (yet) read file format {suffix}')


# ----------------------------------------- TEST:
if __name__ == '__main__':
    from pathlib import Path
    import ema_logging

    work_path = Path.home() / 'Documents' / 'EMA_sim'  # or whatever...
    data_path = work_path / 'data_xlsx' / 'Age_old'  # to use simulation data generated by run_sim.py
    # data_path = work_path / 'data' / 'Age_old'  # to use simulation data generated by run_sim.py

    ema_logging.setup()  # to save the log file

    ema_file = ema_gen(data_path / 'Pop0_S1.xlsx',
                       participant='sheet',
                       dtype={'CoSS': pd.StringDtype()}  # ******** needed ?
                       )
    # for (s, ema) in ema_file:
    for ema in ema_file:
        print(ema.head(100))
        print('dtypes=\n', ema.dtypes)
        count = ema.value_counts(subset=['HA', 'CoSS'], sort=False)
        # ind = pd.MultiIndex.from_frame(ema[['HA', 'CoSS']])
        # c1 = count.reindex(index=ind, fill_value=0)
        print(count)

    path = data_path / 'test_write' / 'test_participant.csv'
    path.parent.mkdir(parents=True, exist_ok=True)
    # ********* check CoSS data type ? should be string ****'
    Table(ema).save(path)



"""Set up logging formats for EmaCalc package"""

import logging
from pathlib import Path


def setup(log_file=None, save_path='.'):
    """create logging handlers
    :param log_file: (optional) string file name for logger FileHandler
    :param save_path: directory where log_file, if any, is stored
    :return: None
    """
    root_logger = logging.getLogger()
    root_logger.setLevel(logging.INFO)
    # root_logger.setLevel(logging.DEBUG)

    # *** create handlers explicitly, to specify encoding:
    if log_file is not None:
        save_path = Path(save_path)
        save_path.mkdir(parents=True, exist_ok=True)  # *** allow over-write ***
        fh = logging.FileHandler(save_path / log_file, mode='w', encoding='utf-8')
        fh.setFormatter(logging.Formatter('{asctime} {name} {levelname}: {message}',
                                          style='{',
                                          datefmt='%Y-%m-%d %H:%M:%S')
                        )
        root_logger.addHandler(fh)

    console = logging.StreamHandler()
    console.setFormatter(logging.Formatter('{asctime} {name}: {message}',
                                           style='{',
                                           datefmt='%H:%M:%S'))
    root_logger.addHandler(console)


"""Help classes for distribution of latent sensory variable
used in EmaCalc.ema_model.EmaModel

*** Version history:
* Version 0.9.2:
2022-06-12, Bradley.log_cdf_diff: check for numerical instability with extreme sample values

* Version 0.7.1:
2022-01-19, methods rvs moved to ema_simulation, only needed there

2021-10-29, copied from PairedCompCalc.pc_model, slightly modified for EmaCalc
2021-11-10, modified Bradley variant for EMA model
2021-11-21, tested Bradley, Thurstone variants for EMA model
"""
import numpy as np
from scipy.special import expit, ndtr
import logging

logger = logging.getLogger(__name__)
# logger.setLevel(logging.DEBUG)  # *** TEST


# ------------------------------------------------------------------
class Bradley:
    """Distribution of latent decision variable in the Bradley-Terry-Luce model,
    with distribution = standard logistic
    cdf(x) = 1 / (1 + exp(-x)) = expit(x)
    """
    unit_label = 'logit unit'
    # for axis label in attribute plots
    scale = np.pi / np.sqrt(3.)
    # = st.dev. of standard logistic distribution
    # may be used to standardize result scale in displays

    @staticmethod
    def log_cdf_diff(a, b):
        """log prob( a < Z <= b)
        where Z is a standard logistic random variable
        :param a: = array with LOWER interval limits
        :param b: = arrays with UPPER interval limits
            a.shape == b.shape
            all( a < b )
        :return: log_p = array with log probabilities, element-wise
            log_p.shape == a.shape == b.shape
        """
        # **** Fix numerical divide by zero in log, asymmetric numerical sensitivity!
        # ***** problem when b = inf and a > 36.7  -> d == 0.
        # ***** No problem with b = -100. and a = -inf
        d = expit(b) - expit(a)
        # *** this may be inaccurately == 0. if 36.7. < a < b in double prec.
        non_positive = d <= 0.
        if np.any(non_positive):
            d[non_positive] = expit(-a[non_positive]) - expit(-b[non_positive])  # using expit symmetry
        non_positive = np.logical_or(d <= 0., np.isnan(d))
        if np.any(non_positive):  # some other error
            logger.warning(f'Bradley.log_cdf_diff: {np.sum(non_positive)} non-positive probability values. '
                           + 'Should never happen!')
            d[non_positive] = np.finfo(float).tiny  # to avoid log(0.)
            # No problem if this logprob multiplied by zero count later
        return np.log(d)

    @staticmethod
    def d_log_cdf_diff(a, b):
        """Element-wise partial derivatives of log_cdf_diff(a, b)
        :param a: = array with LOWER interval limits
        :param b: = arrays with UPPER interval limits
            a.shape == b.shape
            all( a < b )
        :return: tuple (dll_da, dll_db) of arrays, where
            dll_da[...] = d log_cdf_diff(a[...], b[...]) / d a[...]
            dll_db[...] = d log_cdf_diff(a[...], b[...]) / d b[...]
            dll_da.shape == dll_db.shape == a.shape == b.shape
        2017-12-08, tested by finite-diff comparison
        """
        # **** Check for any a - b == 0. or a - b == NaN ? ********
        d = b - a
        error_d = np.logical_or(d == 0., np.isnan(d))
        if np.any(error_d):
            logger.warning(f'Bradley.d_log_cdf_diff: {np.sum(error_d)} non-positive probability values. '
                           + 'Should never happen!')
            d[error_d] = np.finfo(float).tiny  # to avoid expm1(0.) -> 1 / 0.
        # No problem if this logprob multiplied by zero count later
        return (1. / np.expm1(-d) + expit(-a),
                1. / np.expm1(d) + expit(-b))

    def __repr__(self):
        return f'<class {self.__class__.__name__}>'


class Thurstone:
    """Distribution of decision variable in the Thurstone Case V model.
    with distribution function
    cdf(x) = Phi(x)
    NOTE: Thurstone.scale is now included in all calculations.
    This is different from corresponding MatLab code.
    """
    unit_label = 'd-prime unit'

    sqrt_2pi = np.sqrt(2. * np.pi)
    scale = 1.

    @classmethod
    def cdf_diff(cls, a, b):
        """prob( a < Z <= b)
        where Z is a Gaussian standard random variable with variance = 2.
        :param a: = array with LOWER interval limits
        :param b: = arrays with UPPER interval limits
            a.shape == b.shape
            all( a < b )
        :return: log_p = array with log probabilities, element-wise
            log_p.shape == a.shape == b.shape
        2018-04-24, include Thurstone.scale factor
        """
        # **** Check for too small interval, like Bradley ? ************
        a_s = a / cls.scale
        b_s = b / cls.scale
        cdf_diff = ndtr(b_s) - ndtr(a_s)
        ch_sign = a_s >= 0.  # and (b > a) always
        cdf_diff[ch_sign] = ndtr(-a_s[ch_sign]) - ndtr(-b_s[ch_sign])
        # using ndtr symmetry
        # better precision of difference in case of large positive a, b
        return cdf_diff

    @classmethod
    def log_cdf_diff(cls, a, b):
        """log cdf_diff(a, b)
        :param a: = array with LOWER interval limits
        :param b: = arrays with UPPER interval limits
            a.shape == b.shape
            all( a < b )
        :return: log_p = array with log probabilities, element-wise
            log_p.shape == a.shape == b.shape
        """
        return np.log(cls.cdf_diff(a, b))

    @classmethod
    def d_log_cdf_diff(cls, a, b):
        """Element-wise partial derivatives of log_cdf_diff(a, b)
        :param a: = array with LOWER interval limits
        :param b: = arrays with UPPER interval limits
            a.shape == b.shape
            all( a < b )
        :return: tuple (dll_da, dll_db), where
            dll_da[...] = d log_cdf_diff[a[...], b[...]) / d a[...]
            dll_db[...] = d log_cdf_diff[a[...], b[...]) / d b[...]
            dll_da.shape == dll_db.shape == a.shape == b.shape
        Arne Leijon, 2017-12-08, tested by finite-diff comparison
        2018-04-24, include Thurstone.scale factor
        """
        # *** scale may be omitted == 1. anyway
        def norm_pdf(x):
            """Gaussian density function = derivative of ndtr(x / scale)
            """
            return np.exp(- (x / cls.scale)**2 / 2) / cls.sqrt_2pi / cls.scale
        # ----------------------------------------------------
        cdf_diff = cls.cdf_diff(a, b)
        return - norm_pdf(a) / cdf_diff, norm_pdf(b) / cdf_diff

    def __repr__(self):
        return f'<class {self.__class__.__name__}>'


# ------------------------------------------------- TEST:
if __name__ == '__main__':
    from scipy.optimize import approx_fprime, check_grad

    print('*** Testing Bradley and Thurstone gradients')

    # --------------------------------------------------
    for cls in (Bradley, Thurstone):

        print(f'\nTesting {cls}.d_log_cdf_diff() with 1D (a, b) args')
        tau = np.array([0., 2.])
        # tau = np.array([-50.1, -50.])
        # tau = np.array([50., 50.1])
        th = np.array([1.5])
        (a, b) = (tau[0] - th, tau[1] - th)

        def fun_a(da):
            return cls.log_cdf_diff(a + da, b)

        def jac_a(da):
            return cls.d_log_cdf_diff(a + da, b)[0]


        def fun_b(db):
            return cls.log_cdf_diff(a, b + db)

        def jac_b(db):
            return cls.d_log_cdf_diff(a, b + db)[1]

        d = np.array([0.1])
        print(f'a ={a}. b = {b}. d = {d}')
        print(f'{cls}.fun_a_b({d}) = {(fun_a(d), fun_b(d))}')

        print('approx d_log_cdf_diff_da = ', approx_fprime(d, fun_a, epsilon=1e-6))
        print('exact  d_log_cdf_diff_da = ', jac_a(d))
        err = check_grad(fun_a, jac_a, d, epsilon=1e-6)
        print('check_grad err = ', err)

        print('')
        print('approx d_log_cdf_diff_db = ', approx_fprime(d, fun_b, epsilon=1e-6))
        print('exact  d_log_cdf_diff_db = ', jac_b(d))
        err = check_grad(fun_b, jac_b, d, epsilon=1e-6)
        print('check_grad err = ', err)



"""This module defines classes and methods for simulated EMA study
with group(s) of respondents drawn at random from population(s)
with specified inter-individual distributions of all model parameters.

The simulations can generate artificial data with the same structure
as a real experiment.


*** Main Module Classes:

EmaSimPopulation: defines distribution of model parameters
    for probabilities of one or more SITUATIONS, in one or more Situation Dimensions,
    and one or more perceptual ATTRIBUTE latent-variable locations
    in ONE (sub-)population of potential participants.

EmaSimExperiment: defines a complete EMA experiment,
    generates simulated responses by participants in one or more groups,
    sampled from EmaSimPopulation instance(s).
    The EMA data layout is defined by an ema_data.EmaFrame instance,
    defining Situation dimensions and categories, and
    defining discrete ordinal response scales for each Attribute.

EmaSimSubject = superclass for subject properties

SubjectBradley: single subject with Attribute ratings determined by
    the Bradley-Terry-Luce (BTL) model, assuming standard Logistic latent variable,
    i.e., with st.dev. approx= 1.8

SubjectThurstone: single subject with Attribute ratings determined by
    the Thurstone Case V model, assuming standard Gaussian latent variable,
    i.e., with st.dev. = 1.

EmaSimGroup: container for EmaSimSubject instances
    drawn at random from an EmaSimPopulation instance.


*** Main Class Methods:

EmaSimPopulation.gen_group(...)
    draws a group of simulated participants at random from the simulated Population.

EmaSimExperiment.gen_dataset(...)
    generates am ema_data.EmaDataSet instance with simulated EMA records
    for one or more groups of simulated participants.
    All records can be saved to file(s) using the EmaDataSet.save(...) method.

*** Usage example: See script run_sim.py

*** Version History:
* Version 0.9.4:
2022-11-22, allow EmaSimPopulation settings restrict_attribute, restrict_threshold
            with same effects as in ema_model.EmaModel
2022-11-19, adapted to use ema_thresholds.ThresholdsOld

* Version 0.9:
2022-03-21, adapted to use Pandas in EmaFrame and EmaDataSet

* Version 0.7.1:
2022-01-19, module function set_sim_seed to ensure reproducibility
2022-01-19, EmaSimSubject.rvs methods defined here, to ensure reproducibility
2022-01-06, EmaSimPopulation.response_width_mean to control individual decision thresholds
2022-01-05, minor cleanup EmaSimExperiment

* Version 0.5.1:
2021-11-25, allow experiment with NO Attributes

* Version 0.5:
2021-11-14, copied and modified PairedCompCalc -> EmaCalc
2021-11-16, first functional version
2021-11-21, rather extensively tested
"""
# *** store true EmaSimSubject, EmaSimPopulation properties as pd.DataFrame instances ? *****

# **** specify separate Attribute variance for each Situation dimension ????
# **** separate inter-individual attribute variance jointly for all situation_dtypes,
# **** plus variance for independent variations across situation_dtypes???

import numpy as np
import pandas as pd
import logging

from .ema_data import EmaDataSet
# from .ema_thresholds import ThresholdsOld
from .ema_thresholds import ThresholdsFree, ThresholdsMidFixed
from . import ema_latent


RNG = np.random.default_rng()
# = default module-global random generator
# may be modified via function set_sim_seed

logger = logging.getLogger(__name__)
# logger.setLevel(logging.DEBUG)  # test


# ------------------ module functions
def set_sim_seed(seed=None):
    """Set module-global RNG with given seed.
    To be called by user BEFORE any other work,
    to achieve reproducible results.
    :param seed: (optional) integer
    :return: None
    """
    global RNG
    RNG = np.random.default_rng(seed)
    if seed is not None:
        logger.warning(f'*** Using seed={seed} -> reproducible results.')


# --------------------------------------- subject response models
class EmaSimSubject:
    """Simulate one individual participant in an EMA data-collection experiment.
    Superclass for either SubjectBradley or SubjectThurstone
    """
    def __init__(self, sc_prob, a_theta, a_tau):  # lapse_prob=0.):
        """
        :param sc_prob: mD array with (non-normalized) conditional Situation probability
            sc_prob[k0, k1, k2,...] propto Prob (k1, k2, ...)-th situation, GIVEN k0-th Phase
            sc_prob.shape == emf.situation_shape
        :param a_theta: dict with (a_key, a_theta) elements, where
            a_key = a string identifying ONE Attribute among emf.attribute_grades.keys()
            a_theta[k0, k1, ...] = mD array with mean of latent sensory variable for Attribute a_key,
        :param a_tau: dict with (a_key, thr) elements, where
            tnr[l] = UPPER interval limit for l-th ordinal response,
            NOT INCLUDING extreme -inf, +inf limits
        """
        self.sc_prob = sc_prob
        self.a_theta = a_theta
        self.a_tau = a_tau

    def __repr__(self):
        return (f'{self.__class__.__name__}(\n\t' +
                ',\n\t'.join(f'{key}={repr(v)}'
                             for (key, v) in vars(self).items()) +
                '\n\t)')

    def gen_ema_records(self, emf, min_ema, max_ema):
        """Generate a sequence of EMA records at random
        using response properties of self.
        :param emf: ema_data.EmaFrame instance
        :param min_ema: min random number of EMA records per Phase
        :param max_ema: max random number of EMA records per Phase
        :return: ema_df = a pd.DataFrame instance containing simulated EMA results,
            stored according to given emf.
            One column for each Situation dimension and each Attribute.
            One row for each simulated EMA record.
            Number of records randomly drawn with
            min_ema <= ema_df.shape[0] < max_ema
        """
        def situation_index(sit_p, sit_shape):
            """Generate ONE random situation index tuple, NOT including Phase index
            :param sit_p: 1D probability-mass array for all situations, EXCEPT Phase
            :param sit_shape: tuple with situation_shape EXCEPT Phase dimension
            :return: ind = index tuple; len(ind) = len(sc_shape)
             """
            sit_ind = RNG.choice(len(sit_p), p=sit_p)  # linear random index
            return np.unravel_index(sit_ind, shape=sit_shape)

        def situation_dict(sit_ind):
            """Convert index tuple to situation dict
            """
            return {sit_key: sit_dtype.categories[i]
                    for (i, (sit_key, sit_dtype)) in zip(sit_ind,
                                                         emf.situation_dtypes.items())}
        # -------------------------------------------------------

        ema_list = []
        # = list of dicts, one for each simulated EMA record
        sit_shape_phase = emf.situation_shape[1:]
        sit_prob_phase = self.sc_prob.reshape((self.sc_prob.shape[0], -1))
        for (i, p) in enumerate(sit_prob_phase):
            n_rec = RNG.integers(min_ema, max_ema)
            for _ in range(n_rec):
                sit_index = (i, *situation_index(p, sit_shape_phase))
                a_grades = self.gen_attr_grades(emf, sit_index)
                ema_list.append(situation_dict(sit_index) | a_grades)
        cat_dtypes = emf.situation_dtypes | emf.attribute_dtypes  # ****** needed ?
        return pd.DataFrame.from_records(ema_list).astype(cat_dtypes)

    def gen_attr_grades(self, emf, sit_index):
        """Generate random ordinal Attribute grades in given Situation
        :param emf: external ema_data.EmaFrame object defining experimental layout
        :param sit_index: index tuple defining ONE situation for this record,
            including index in ALL situation dimensions
        :return: a_grades = dict with elements (a, grade)
        """
        def rvs_grade(th, tau):
            """Generate a random grade index for ONE attribute in ONE situation
            :param th: scalar location of latent variable in the given situation
            :param tau: 1D threshold array for this attribute,
                NOT INCLUDING extreme -inf, +inf limits
            :return: scalar integer index of given rating
            """
            x = self.rvs(th)  # done by sub-class
            return np.sum(x > tau)
        # -------------------------------------------------------
        return {a: emf.attribute_dtypes[a].categories[rvs_grade(a_th[sit_index],
                                                                self.a_tau[a])]
                for (a, a_th) in self.a_theta.items()}

    @staticmethod
    def rvs(loc, size=None):
        """Abstract method, implemented by sub-class.
        Draw random variables from self
         :param loc: scalar or array-like location
         :param size: (optional) size of result
         :return: x = scalar or array
             x.shape == loc.shape, if no size is given
         """
        raise NotImplementedError

    # def lapse(self):  # *** future ???
    #     """Generate True with prob = self.lapse_prob
    #     """
    #     return uniform.rvs(0, 1.) < self.lapse_prob
    #
    # def lapse_response(self):
    #     """Generate a random result, disregarding latent variable and threshold parameters
    #     :return: scalar integer
    #         in {-n_difference_grades, ...,-1, +1,..., + n_difference_grades}, if forced_choice
    #         in {-n_difference_grades+1, ...,0, ...,  + n_difference_grades-1}, if not forced_choice
    #         i.e., excluding 0 if self.emf.forced_choice
    #     """
    #     n_response_limits = len(self.response_thresholds)
    #     # if self.emf.forced_choice:
    #     if self.response_thresholds[0] == 0.:  # forced_choice
    #         return ((-1 if uniform.rvs() < 0.5 else +1) *
    #                 randint.rvs(low=1, high=n_response_limits + 1))
    #
    #     else:
    #         return randint.rvs(low=-n_response_limits,
    #                            high=n_response_limits + 1)
    #


class SubjectThurstone(ema_latent.Thurstone, EmaSimSubject):
    """Simulate a subject using the Thurstone Case V choice model.
    """
    @staticmethod
    def rvs(loc, size=None):
        """Draw random sample(s) from self
        :param loc: scalar or array-like location
        :param size: (optional) size of result
        :return: x = scalar or array
            x.shape == loc.shape, if no size is given
        """
        loc = np.asarray(loc)
        if size is None:
            size = loc.shape
        return loc + RNG.standard_normal(size=size)


class SubjectBradley(ema_latent.Bradley, EmaSimSubject):
    """Simulate one individual participant in a paired-comparison experiment.
    The subject responds using the Bradley-Terry-Luce choice model,
    with parameters defined in the log domain.
    """
    @staticmethod
    def rvs(loc, size=None):
        """Draw random variables from self
        :param loc: scalar or array-like location
        :param size: (optional) size of result
        :return: x = scalar or array
            x.shape == loc.shape
        """
        return RNG.logistic(loc, size=size)


# -------------------------------------------------------------------------
class EmaSimPopulation:
    """Defines a simulated population
    from which groups of random test participants can be generated
    for a simulated EMA experiment.

    The population instance defines distributions for
    one or more nominal Situation categories, and
    zero, one or more perceptual Attributes, given any Situation.

    Each Situation is a combination of one category from each Situation Dimension.
    """
    def __init__(self, emf,
                 situation_prob,
                 attribute_mean=None,
                 log_situation_std=0.,
                 attribute_std=0.,
                 response_width_mean=None,  # -> subject_class.scale
                 log_response_width_std=0.,
                 lapse_prob_range=(0., 0.),  # ******* not used ******
                 subject_class=SubjectBradley,
                 restrict_attribute=False,
                 restrict_threshold=True,
                 id=''):
        """
        :param emf: ema_data.EmaFrame instance defining Situations and Attributes
        :param situation_prob: array-like multi-dim, with
            situation_prob[k0, k1,...] prop.to probability of (k1, k2,...)-th situation,
            GIVEN the k0-th test phase category, even if only one phase category.
            situation_prob.shape must correspond to shape of emf.situation_dtypes.
        :param attribute_mean: (optional) dict or iterable with elements (a_key, a_mean), where
            a_key is string id of a rated perceptual attribute,
            a_mean is an mD array with
            a_mean[k0, k1, ...] = latent-variable mean for attribute a_key, in subject_class scale units,
            given the (k0, k1,...)-th Situation category, i.e.,
            a_mean.shape == situation_prob.shape, for all attributes.
        :param log_situation_std: (optional) inter-individual standard deviation of
            log probabilities for each situation category.
        :param attribute_std: (optional) scalar inter-individual standard deviation of all attribute parameters
        :param response_width_mean: (optional) mean width of response intervals,
            in subject_class scale units
        :param log_response_width_std: (optional) scalar standard deviation of log(response-interval-width)
        :param lapse_prob_range: (optional) tuple (min, max) probability of random lapse response
        :param subject_class: (optional) subject probabilistic model for generating responses
        :param restrict_attribute: (optional) boolean switch
            to force mean attribute location == 0., mean across situations
        :param restrict_threshold: (optional) boolean switch
            to force ONE mid-range response-threshold == 0.
        :param id: (optional) string label, used as prefix in all generated subject names
        """
        self.emf = emf  # save emf ref here, too, although same all sub-populations
        situation_prob = np.asarray(situation_prob)
        if emf.situation_shape[0] == 1 and emf.situation_shape[1:] == situation_prob.shape:
            situation_prob = situation_prob[None, ...]
        if emf.situation_shape != situation_prob.shape:
            raise RuntimeError('situation_prob.shape must agree with EmaFrame situation_dtypes')
        self.situation_prob = situation_prob
        for (i, sc_i) in enumerate(self.situation_prob):
            self.situation_prob[i] /= np.sum(sc_i)
            # = normalizes conditional probabilities, for each phase
        self.log_situation_std = log_situation_std

        if attribute_mean is None:
            attribute_mean = dict()  # NO attributes
        else:
            attribute_mean = dict(attribute_mean)
        if set(attribute_mean.keys()) != set(emf.attribute_dtypes.keys()):
            raise RuntimeError('attribute_mean must define same attributes as EmaFrame')
        for (a, a_mean) in attribute_mean.items():
            if (emf.situation_shape[0] == 1
                    and emf.situation_shape[1:] == a_mean.shape):
                attribute_mean[a] = attribute_mean[a][None, ...]
        assert all(a_mean.shape == emf.situation_shape
                   for a_mean in attribute_mean.values()), 'attribute_mean.shape must match situation_shape'
        if restrict_attribute and emf.tied_response_scales:
            restrict_attribute = False
            logger.warning('*** Cannot use restrict_attribute=True with tied response scales')
        if restrict_attribute and restrict_threshold:
            restrict_attribute = False  # ONLY ONE restriction allowed
            logger.warning(f'Only ONE restriction allowed: using restrict_threshold={restrict_threshold}')
        self.restrict_attribute = restrict_attribute
        self.restrict_threshold = restrict_threshold
        if self.restrict_attribute:
            attribute_mean = {a: a_mean - np.mean(a_mean)
                              for (a, a_mean) in attribute_mean.items()}
        self.attribute_mean = attribute_mean
        self.attribute_std = attribute_std
        if response_width_mean is None:
            response_width_mean = subject_class.scale
        self.response_width_mean = response_width_mean
        # if self.restrict_threshold:
        #     self._thr = ThresholdsMidFixed
        # else:
        #     self._thr = ThresholdsFree
        # self._eta = _init_eta(response_width_mean, self._thr)
        self.log_response_width_std = log_response_width_std
        # = inter-individual std around self._eta
        self.lapse_prob_range = lapse_prob_range
        self.subject_class = subject_class
        self.id = id

    def __repr__(self):
        return (f'{self.__class__.__name__}(\n\t' +
                ',\n\t'.join(f'{key}={repr(v)}'
                             for (key, v) in vars(self).items()) +
                '\n\t)')

    @property
    def n_attributes(self):
        return len(self.attribute_mean)

    def gen_group(self, n_participants=1):
        """Create a group of simulated-subject instances randomly drawn from self,
        with properties suitable for a planned experiment.
        :param n_participants: number of randomly drawn participants from self
        :return: single EmaSimGroup instance, containing generated participants,
            each with properties drawn from self.
        """
        def gen_sc_prob():
            sc_p = np.exp(np.log(self.situation_prob) +
                          self.log_situation_std *
                          RNG.standard_normal(size=self.emf.situation_shape))
            for (i, sc_i) in enumerate(sc_p):
                sc_p[i] /= np.sum(sc_i)
                # normalized conditional probabilities, for each phase
            return sc_p

        def gen_attr():
            theta = self.attribute_std * RNG.standard_normal(size=(self.n_attributes,
                                                                   *self.emf.situation_shape))
            # theta[i, ...] = random offset of location for i-th attribute of s-th subject
            return {a: th_mean + th_d
                    for ((a, th_mean), th_d) in zip(self.attribute_mean.items(),
                                                    theta)}

        def gen_tau():
            """Random threshold parameters for each attribute
            """
            def tau(n_levels, thr):
                """Calc response thresholds
                :param n_levels: integer number of response intervals
                :param thr: Thresholds subclass, mapping parameters eta -> thresholds
                :return: t = 1D array with INNER interval limits, i.e.,
                    t[m] = UPPER interval limit for m-th ordinal response
                    t.shape == (n_levels - 1,)
                """
                t = self.response_width_mean * (np.arange(n_levels - 1) + 1. - n_levels / 2)
                mean_tau = thr.tau(thr.tau_inv(t))  # ****** test
                n_eta = thr.n_param(n_levels)
                eta = thr.tau_inv(t) + self.log_response_width_std * RNG.standard_normal(size=n_eta)
                # incl. random variations of log interval width in mapped [0, 1] range
                return thr.tau(eta)[1:-1]  # EXCL (-inf, +inf) extremes
            # ----------------------------------------------------
            if self.restrict_threshold:
                thr = ThresholdsMidFixed
            else:
                thr = ThresholdsFree
            scale_tau = {s_id: tau(len(scale.categories), thr)
                         for (s_id, scale) in self.emf.ordinal_scales.items()}
            # = scale thresholds, possibly used for more than one attribute
            return {a: scale_tau[s_id]
                    for (a, s_id) in self.emf.attribute_scales.items()}
            # --------------------------------------------------------------------
        return EmaSimGroup(self,
                           {self.id + f'_S{i}': self.subject_class(sc_prob=gen_sc_prob(),
                                                                   a_theta=gen_attr(),
                                                                   a_tau=gen_tau())
                            for i in range(n_participants)})


# ----------------------------------------------------------------------
class EmaSimGroup:
    """Group of test participants drawn from a given population
    """
    def __init__(self, pop, participants):
        """
        :param pop: an EmaSimPopulation instance,
            from which participants have been drawn at random
        :param participants: dict with items (s_id, s_sim)
            s_id = string id for the subject
            s_sim = an EmaSimSubject instance
        """
        self.pop = pop
        self.participants = participants

    def __repr__(self):
        # n_ema = sum(s_df.shape[0] for s_df in self.participants.values())
        return (self.__class__.__name__ + '('
                + f'\n\tpop={repr(self.pop)}'
                + f'\n\tparticipants= dict with {len(self.participants)} simulated participants)')


# -------------------------------------------------------------------------
class EmaSimExperiment:
    """Defines a simulated EMA data-collection experiment,
    with one or more groups of simulated participants, with
    each group generated from an EmaSimPopulation instance.

    Method gen_dataset() generates a complete ema_data.EmaDataSet instance
    with EMA records for all participants in all groups.

    The experimental procedure is defined by
    emf = an ema_data.EmaFrame instance
    """
    def __init__(self, emf, groups):
        """
        :param emf: EmaFrame instance, defining experimental parameters
        :param groups: dict with elements (g_id, g_sim),
            g_id is a tuple with one or several group-category labels,
                one for each element in emf.group_head()
            g_sim = an EmaSimGroup instance
        :param restrict_attribute: (optional) boolean switch
            to force restriction on attribute sensory-variable locations
        :param restrict_threshold: (optional) boolean switch
            to force restriction on response-threshold locations
        """
        self.emf = emf
        self.groups = groups

    def __repr__(self):
        return (f'{self.__class__.__name__}(\n\t' +
                ',\n\t'.join(f'{key}={repr(v)}'
                             for (key, v) in vars(self).items()) +
                '\n\t)')

    def gen_dataset(self, min_ema=3, max_ema=50):
        """Generate a complete EmaDataSet instance for this experiment,
        with one or more groups of participants.
        :param min_ema: min random number of EMA records in each Phase
        :param max_ema: max random number of EMA records in each Phase
        :return: a single EmaDataSet instance
        """
        emd = {g: {s_id: s.gen_ema_records(self.emf, min_ema, max_ema)
                   for (s_id, s) in g_sim.participants.items()}
               for (g, g_sim) in self.groups.items()}
        return EmaDataSet(self.emf, emd)
